#include <config.h>			/* private file */

/*
 * regular expression matching routines for tgrind/tfontedpr.
 *
 * These routines were written by Dave Presotto (I think) for vgrind.
 * Minor mods & attempts to improve performance by Van Jacobson (van@lbl-rtsg)
 * and Chris Torek (chris@maryland).
 *
 * Modifications.
 * --------------
 * 30Mar85 Van & Chris	Changed expmatch to return pointer to start of what
 *			was matched in addition to pointer to match end.
 *			Several changes to improve performance (too numerous
 *	expmatch		to mention).
 * 11Dec84 Dave Presotto  Written.
 */

#if defined(HAVE_CTYPE_H)
#include <ctype.h>
#endif

#if defined(HAVE_STDIO_H)
#include <stdio.h>
#endif

#if defined(HAVE_STDLIB_H)
#include <stdlib.h>
#endif

#if defined(HAVE_STRING_H)
#include <string.h>
#endif

#include <stdc.h>			/* private file */
#include <boolean.h>			/* private file */
#include <regexp.h>			/* private file */

#define CHRVAL(c)	(0xff & (c))
#define NIL		((char*)NULL)
#define NUL		'\0'

#define elementsof(v)	(sizeof(v)/sizeof(v[0]))
#define makelower(c)	(isupper((c)) ? tolower((c)) : (c))

boolean isidchar[256];		/* indexed by unsigned characters: TRUE if */
				/* the character can occur in a \p or keyword */
				/* string */

boolean isinichar[256];		/* indexed by unsigned characters: TRUE if */
				/* the character is a legal initial character */
				/* of a procedure name or keyword */

static void	chklength ARGS((int n_));

/*  lc_strncmp - like strncmp except that we (implicitly) convert the
 *	 	first string to lower case before comparing.
 */


#if STDC
int
lc_strncmp(register const char *s1, register const char *s2, register size_t len)
#else
lc_strncmp(s1, s2, len)
	register const char *s1;
	register const char *s2;
	register size_t len;
#endif
{
	while (len-- > 0)
	    if (*s2 - makelower(*s1))
		    return(1);
	    else
		s2++, s1++;

	return(0);
}

/*	The following routine converts an irregular expression to
 *	internal format.
 *
 *	Either meta symbols (\a \d or \p) or character strings or
 *	operations ( alternation or perenthesizing ) can be
 *	specified.  Each starts with a descriptor byte.  The descriptor
 *	byte has STR set for strings, META set for meta symbols
 *	and OPER set for operations.
 *	The descriptor byte can also have the OPT bit set if the object
 *	defined is optional.  Also ALT can be set to indicate an alternation.
 *
 *	For metasymbols the byte following the descriptor byte identities
 *	the meta symbol (containing an ascii 'a', 'd', 'p', '|', or '(').  For
 *	strings the byte after the descriptor is a character count for
 *	the string:
 *
 *		meta symbols := descriptor
 *				symbol
 *
 *		strings :=	descriptor
 *				character count
 *				the string
 *
 *		operations :=	descriptor
 *				symbol
 *				character count
 */

/*
 *  handy macros for accessing parts of match blocks
 */
#define MSYM(A) (*(A+1))		/* symbol in a meta symbol block */
#define MNEXT(A) (A+2)			/* address of the character following a metasymbol block */

#define OSYM(A) (*(A+1))		/* symbol in an operation block */
#define OCNT(A) (*((unsigned char*)A+2)) /* character count */
#define ONEXT(A) (A+3)			/* address of the next character after the operation */
#define OPTR(A) (A+*((unsigned const char*)A+2)) /* place pointed to by the operator */

#define SCNT(A) (*((unsigned char*)A+1)) /* byte count of a string */
#define SSTR(A) (A+2)			/* address of the string */
#define SNEXT(A) (A+2+*((unsigned const char*)A+1))	/* address of the character following the string */

/*
 *  bit flags in the descriptor
 */
#define OPT 1
#define STR 2
#define META 4
#define ALT 8
#define OPER 16

const char *ure;	/* pointer current position in unconverted exp */
char *ccre;		/* pointer to current position in converted exp*/

#if STDC
char *
convexp(const char *re)
             		/* unconverted irregular expression */
#else
char *
convexp(re)
    const char *re;	/* unconverted irregular expression */
#endif
{
    register char *cre;		/* pointer to converted regular expression */

    /* allocate room for the converted expression */
    if (re == NIL)
	return (NIL);
    if (*re == NUL)
	return (NIL);
    cre = (char*)malloc (4 * strlen(re) + 3);
    ccre = cre;
    ure = re;

    /* start the conversion with a \a */
    *cre = META | OPT;
    MSYM(cre) = 'a';
    ccre = MNEXT(cre);

    /* start the conversion (it's recursive) */
    expconv ();
    *ccre = NUL;
    return (cre);
}

void
expconv(VOID_ARG)
{
    register char *cs;		/* pointer to current symbol in converted exp */
    register char c;		/* character being processed */
    register char *acs;		/* pointer to last alternate */
    register int temp;

    /* let the conversion begin */
    acs = NIL;
    cs = NIL;
    while (*ure != NUL) {
	switch (c = *ure++) {

	case '\\':
	    switch (c = *ure++) {

	    /* escaped characters are just characters */
	    default:
		if (cs == NIL || (*cs & STR) == 0) {
		    cs = ccre;
		    *cs = STR;
		    SCNT(cs) = 1;
		    ccre += 2;
		} else
		    SCNT(cs)++;
		*ccre++ = c;
		break;

	    /* normal(?) metacharacters */
	    case 'a':
	    case 'd':
	    case 'e':
	    case 'p':
		if (acs != NIL && acs != cs) {
		    do {
			temp = OCNT(acs);
			chklength((int)(ccre - acs));
			OCNT(acs) = ccre - acs;
			acs -= temp;
		    } while (temp != 0);
		    acs = NIL;
		}
		cs = ccre;
		*cs = META;
		MSYM(cs) = c;
		ccre = MNEXT(cs);
		break;
	    }
	    break;

	/* just put the symbol in */
	case '^':
	case '$':
	    if (acs != NIL && acs != cs) {
		do {
		    temp = OCNT(acs);
		    chklength((int)(ccre - acs));
		    OCNT(acs) = ccre - acs;
		    acs -= temp;
		} while (temp != 0);
		acs = NIL;
	    }
	    cs = ccre;
	    *cs = META;
	    MSYM(cs) = c;
	    ccre = MNEXT(cs);
	    break;

	/* mark the last match sequence as optional */
	case '?':
	    if (cs)
	    	*cs = *cs | OPT;
	    break;

	/* recurse and define a subexpression */
	case '(':
	    if (acs != NIL && acs != cs) {
		do {
		    temp = OCNT(acs);
		    chklength((int)(ccre - acs));
		    OCNT(acs) = ccre - acs;
		    acs -= temp;
		} while (temp != 0);
		acs = NIL;
	    }
	    cs = ccre;
	    *cs = OPER;
	    OSYM(cs) = '(';
	    ccre = ONEXT(cs);
	    expconv ();
	    chklength((int)(ccre - cs));
	    OCNT(cs) = ccre - cs;		/* offset to next symbol */
	    break;

	/* return from a recursion */
	case ')':
	    if (acs != NIL) {
		do {
		    temp = OCNT(acs);
		    chklength((int)(ccre - acs));
		    OCNT(acs) = ccre - acs;
		    acs -= temp;
		} while (temp != 0);
		acs = NIL;
	    }
	    cs = ccre;
	    *cs = META;
	    MSYM(cs) = c;
	    ccre = MNEXT(cs);
	    return;

	/* mark the last match sequence as having an alternate */
	/* the third byte will contain an offset to jump over the */
	/* alternate match in case the first did not fail */
	case '|':
	    if (acs != NIL && acs != cs)
	    {
		chklength((int)(ccre - acs));
		OCNT(ccre) = ccre - acs;	/* make a back pointer */
	    }
	    else
		OCNT(ccre) = 0;
	    *cs |= ALT;
	    cs = ccre;
	    *cs = OPER;
	    OSYM(cs) = '|';
	    ccre = ONEXT(cs);
	    acs = cs;	/* remember that the pointer is to be filles */
	    break;

	/* if its not a metasymbol just build a scharacter string */
	default:
	    if (cs == NIL || (*cs & STR) == 0) {
		cs = ccre;
		*cs = STR;
		SCNT(cs) = 1;
		ccre = SSTR(cs);
	    } else
		SCNT(cs)++;
	    *ccre++ = c;
	    break;
	}
    }
    if (acs != NIL) {
	do {
	    temp = OCNT(acs);
	    chklength((int)(ccre - acs));
	    OCNT(acs) = ccre - acs;
	    acs -= temp;
	} while (temp != 0);
	acs = NIL;
    }
    return;
}


#if STDC
static void
chklength(int n)
#else
static void
chklength(n)
int n;
#endif
{
    if ((n < 0) || (n > 255))
    {
	(void)fprintf(stderr,
		      "FATAL ERROR: Pattern string length [%d] out of range 0..255.\n", n);
	(void)fprintf(stderr,
		      "Internal data structures in regexp.c require one-byte lengths, so \n");
	(void)fprintf(stderr,
		      "you will have to simplify the pattern containing [...%s].\n",ure);
	exit(EXIT_FAILURE);
    }
}

/* end of convertre */


/*
 *	The following routine recognises an irregular expresion
 *	with the following special characters:
 *
 *		?	-	means last match was optional
 *		\a	-	matches any number of characters
 *		\d	-	matches any number of spaces and tabs
 *		\p	-	matches any number of alphanumeric
 *				characters. The
 *				characters matched will be copied into
 *				the area pointed to by 'name'.
 *		|	-	alternation
 *		( )	-	grouping used mostly for alternation and
 *				optionality
 *
 *	The irregular expression must be translated to internal form
 *	prior to calling this routine
 *
 *	The value returned is the pointer to the last character matched.
 *	If strtptr is non-null, a pointer to the start of the string matched
 *	(excluding \a matches) will be returned at *strtptr.
 *	If mstring is non-null, the string matched by a \p will be copied
 *	into mstring.
 */

#if defined(DEBUG)

static int rlevel = 0;

#define DEBUG_RETURN(number,result) \
	(void)(indent(), \
	       (result == NIL) ? \
	       (void)fprintf(stderr, \
			     "match failure [%d] looking at [%s]\n", \
			     (int)(number), s) : \
	       (void)fprintf(stderr, \
			     "match success [%d] looking at [%s] matched [%s]\n", \
			     (int)(number), s, result) )

void
indent(VOID_ARG)
{
    int k;

    for (k = 0; k < (rlevel - 1); ++k)
	(void)fprintf(stderr,"...");
}



#if STDC
void
debug(register const char *s, register const char *re)
                     		/* string to check for a match in */
                      		/* a converted irregular expression */
#else
void
debug(s, re)
    register const char *s;	/* string to check for a match in */
    register const char *re;	/* a converted irregular expression */
#endif
{
    int k;
    static int first = 1;

    if (first)
    {
        indent();
        (void)fprintf(stderr,"debug: idchar   = [");
        for (k = 0; k < elementsof(isidchar); ++k)
        {
	    if (isidchar[k])
	    {
		if (isprint(k))
    	            (void)fputc(k,stderr);
		else
    	            (void)fprintf(stderr,"<0x%02x>",k);
	    }
        }
	(void)fprintf(stderr,"]\n");

	indent();
	(void)fprintf(stderr,"debug: inichar  = [");
	for (k = 0; k < elementsof(isinichar); ++k)
        {
	    if (isinichar[k])
	    {
		if (isprint(k))
    	            (void)fputc(k,stderr);
		else
    	            (void)fprintf(stderr,"<0x%02x>",k);
	    }
        }
	(void)fprintf(stderr,"]\n");
	first = 0;
    }

    indent();
    (void)fprintf(stderr,"debug: s  = [");
    for (; s && *s; ++s)
    {
	if (isprint(*s))
	    (void)fputc(*s,stderr);
	else
	    (void)fprintf(stderr,"<0x%02x>",*s);
    }
    (void)fputc(']',stderr);
    (void)fputc('\n',stderr);

    indent();
    (void)fprintf(stderr,"debug: re = [");
    for (; re && *re; ++re)
    {
	if (isprint(*re))
	    (void)fputc(*re,stderr);
	else if ((*re != '\n') && (*re < ' ') && (*re >= 0))
	{
	    (void)fprintf(stderr,"<");
	    if (*re & OPT) 	    (void)fprintf(stderr,"OPT_");
	    if (*re & STR) 	    (void)fprintf(stderr,"STR_");
	    if (*re & META) 	    (void)fprintf(stderr,"META_");
	    if (*re & ALT) 	    (void)fprintf(stderr,"ALT_");
	    if (*re & OPER) 	    (void)fprintf(stderr,"OPER_");
	    (void)fprintf(stderr,">");
	}
	else
	    (void)fprintf(stderr,"<0x%02x>",*re);
	if ((*re < ' ') && (*re & STR))
	    (void)fprintf(stderr,"<L%d>",(int)(*(++re)));
    }
    (void)fputc(']',stderr);
    (void)fputc('\n',stderr);
}

const char	*expmatch_ ARGS((const char *s_, const char *re_,
				 const char **strtptr_, char *mstring_));

#if STDC
const char *
expmatch (register const char *s, register const char *re, const char **strtptr,
	  char *mstring)
                     		/* string to check for a match in */
                      		/* a converted irregular expression */
                   		/* where to put ptr to start of match */
                  		/* where to put whatever matches a \p */
#else
const char *
expmatch (s, re, strtptr, mstring)
    register const char *s;	/* string to check for a match in */
    register const char *re;	/* a converted irregular expression */
    const char **strtptr;	/* where to put ptr to start of match */
    char *mstring;		/* where to put whatever matches a \p */
#endif
{
    const char *result;
    int n;

    rlevel++;

    debug(s, re);

    if (mstring)
	*mstring = NUL;
    result = expmatch_(s, re, strtptr, mstring);

    indent();
    (void)fprintf(stderr, "Rest = [%s]", result ? result : "");

    if (strtptr && *strtptr)
    {
	n = (int)(result + 1 - *strtptr);
	if ((n > 0) && (n < 256))
	    (void)fprintf(stderr, "\tMatch = [%.*s]", n, *strtptr);
    }

    if (mstring && *mstring)
	(void)fprintf(stderr, "\tProcname = [%s]", mstring);

    (void)fprintf(stderr, "\n");

    rlevel--;

    return (result);
}

#else /* !defined(DEBUG) */
#define DEBUG_RETURN(number,result)
#define expmatch_ expmatch
#endif /* defined(DEBUG) */


boolean _escaped;		/* true if we are currently _escaped (MUST be
				   set to FALSE by caller of expmatch())*/
const char *Str_Start;		/* start of string (MUST be set by caller
				   (other than expmatch()) to initial s
				   passed to expmatch()) */

int	(*re_strncmp) ARGS((const char *s1_, const char *s2_, size_t len_));

#if STDC
const char *
expmatch_ (register const char *s, register const char *re, const char **strtptr,
	   char *mstring)
                     		/* string to check for a match in */
                      		/* a converted irregular expression */
                   		/* where to put ptr to start of match */
                  		/* where to put whatever matches a \p */
#else
const char *
expmatch_ (s, re, strtptr, mstring)
    register const char *s;	/* string to check for a match in */
    register const char *re;	/* a converted irregular expression */
    char **strtptr;		/* where to put ptr to start of match */
    char *mstring;		/* where to put whatever matches a \p */
#endif
{
    register const char *cs;	/* the current symbol */
    register const char *ptr;	/* temporary pointer */
    register const char *s1;	/* temporary pointer */
    register char c;		/* temporary */
    boolean matched;		/* a temporary boolean */

    /* initial conditions */
    if ( strtptr )
	*strtptr = NIL;
    if (re == NIL)
    {
	DEBUG_RETURN(0,NIL);
	return (NIL);
    }
    cs = re;
    matched = FALSE;

    /* loop till expression string is exhausted (or at least pretty tired) */
    while (*cs) {
	switch (*cs & (OPER | STR | META)) {

	/* try to match a string */
	case STR:
	    matched = ((*re_strncmp)(s, SSTR(cs), SCNT(cs)) == 0) ? TRUE : FALSE;
	    if (matched) {

		/* hoorah it matches */
		s += SCNT(cs);
		cs = SNEXT(cs);
	    } else if (*cs & ALT) {

		/* alternation, skip to next expression */
		cs = SNEXT(cs);
	    } else if (*cs & OPT) {

		/* the match is optional */
		cs = SNEXT(cs);
		matched = TRUE;		/* indicate a successful match */
	    } else {
		DEBUG_RETURN(1,NIL);
		/* no match, error return */
		return (NIL);
	    }
	    break;

	/* an operator, do something fancy */
	case OPER:
	    switch (OSYM(cs)) {

	    /* this is an alternation */
	    case '|':
		if (matched)

		    /* last thing in the alternation was a match, skip ahead */
		    cs = OPTR(cs);
		else

		    /* no match, keep trying */
		    cs = ONEXT(cs);
		break;

	    /* this is a grouping, recurse */
	    case '(':
		ptr = expmatch (s, ONEXT(cs), strtptr, mstring);
		if (ptr != NIL) {

		    /* the subexpression matched */
		    matched = TRUE;
		    s = ptr;
		} else if (*cs & ALT) {

		    /* alternation, skip to next expression */
		    matched = FALSE;
		} else if (*cs & OPT) {

		    /* the match is optional */
		    matched = TRUE;	/* indicate a successful match */
		} else {
		    DEBUG_RETURN(2,NIL);
		    /* no match, error return */
		    return (NIL);
		}
		cs = OPTR(cs);
		break;
	    }
	    break;

	/* try to match a metasymbol */
	case META:
	    switch (MSYM(cs)) {

	    /* try to match anything and remember what was matched */
	    case 'p':
		/*
		 *  This is really the same as trying to match the
		 *  remaining parts of the expression to any subset
		 *  of the string.
		 */
		s1 = s;
		do {
		    ptr = expmatch (s1, MNEXT(cs), strtptr, mstring);
		    if (ptr != NIL && s1 != s) {

			/* we have a match, remember the match */
			if ( mstring ) {
			    strncpy (mstring, s, s1 - s);
			    mstring[s1 - s] = NUL;
			}
			DEBUG_RETURN(102,ptr);
			return (ptr);

		    } else if (ptr != NIL && (*cs & OPT)) {

			/* \p was optional so no match is ok */
			DEBUG_RETURN(103,ptr);
			return (ptr);

		    } else if (ptr != NIL) {
			DEBUG_RETURN(3,NIL);
			/* not optional and we still matched */
			return (NIL);
		    }
#if 0	/* old hardwired test: \p matches [A-Za-z0-9_] */

		    if (!isalnum(*s1) && *s1 != '_')
			return (NIL);
#else	/* new dynamic test: \p matches TRUE entries of isidchar[] */
		    if (s1 == s) /* then initial character of name */
		    {
			if (!isinichar[CHRVAL(*s1)])
			{
			    DEBUG_RETURN(4,NIL);
			    return (NIL);
			}
		    }
		    else	/* non-initial character of name */
		    {
			if (!isidchar[CHRVAL(*s1)])
			{
			    DEBUG_RETURN(5,NIL);
			    return (NIL);
			}
		    }
#endif
		    if (*s1 == '\\')
			_escaped = _escaped ? FALSE : TRUE;
		    else
			_escaped = FALSE;
		} while (*s1++);
		DEBUG_RETURN(6,NIL);
		return (NIL);

	    /* try to match anything */
	    case 'a':
		/*
		 *  This is really the same as trying to match the
		 *  remaining parts of the expression to any subset
		 *  of the string.
		 */
		s1 = s;
		do {
		    /*
		     * Hack for an important special case:  if the next thing
		     * in the pattern is a string, just gobble characters until
		     * we find something that matches that string (this saves
		     * the cost of a recursive call on expmatch while scanning
		     * for the start of comments or strings).  Since many
		     * patterns end with a string, we also treat that as a
		     * special case.
		     */
		    if( *(ptr=MNEXT(cs)) == STR ) {
			c = *SSTR(ptr);
			while( *s1 && *s1 != c )
			    s1++;

			if ( *s1 == NUL )
			{
			    DEBUG_RETURN(7,NIL);
			    return(NIL);
			}

			if ( SNEXT(ptr) == 0 && (s1 != s || *cs & OPT)) {
			    /* next item is a string, it's the last item and
			     * the \a match is ok - just loop to try & match
			     * the string.
			     */
			    if ( strtptr )
				*strtptr = s1;

			    cs = ptr;
			    s = s1;
			    break;
			}
		    }
		    ptr = expmatch (s1, MNEXT(cs), strtptr, mstring);
		    if (ptr != NIL && (s1 != s || *cs & OPT)) {

			/* we have a match */
			if ( strtptr )
			    *strtptr = s1;

			DEBUG_RETURN(104,ptr);
			return(ptr);

		    } else if (ptr != NIL) {
			DEBUG_RETURN(8,NIL);
			/* not optional and we still matched */
			return (NIL);
		    }
		    if (*s1 == '\\')
			_escaped = _escaped ? FALSE : TRUE;
		    else
			_escaped = FALSE;
		} while (*s1++);
		DEBUG_RETURN(9,NIL);
		return (NIL);

	    /* fail if we are currently _escaped */
	    case 'e':
		if (_escaped)
		{
		    DEBUG_RETURN(10,NIL);
		    return(NIL);
		}
		cs = MNEXT(cs);
		break;

	    /* match any number (> 0) of tabs and spaces, or newline, or
	       end of string */
	    case 'd':
		ptr = s;
		while (*s == ' ' || *s == '\t')
		    s++;
		if (s != ptr || s == Str_Start) {

		    /* match, be happy */
		    matched = TRUE;
		    cs = MNEXT(cs);
		} else if (*s == '\n' || *s == NUL) {

		    /* match, be happy */
		    matched = TRUE;
		    cs = MNEXT(cs);
		} else if (*cs & ALT) {

		    /* try the next part */
		    matched = FALSE;
		    cs = MNEXT(cs);
		} else if (*cs & OPT) {

		    /* doesn't matter */
		    matched = TRUE;
		    cs = MNEXT(cs);
		} else {
		    DEBUG_RETURN(11,NIL);
		    /* no match, error return */
		    return (NIL);
		}
		break;

	    /* check for end of line, or end of string */
	    case '$':
		if (*s == NUL || *s == '\n') {

		    /* match, be happy */
		    s++;
		    matched = TRUE;
		    cs = MNEXT(cs);
		} else if (*cs & ALT) {

		    /* try the next part */
		    matched = FALSE;
		    cs = MNEXT(cs);
		} else if (*cs & OPT) {

		    /* doesn't matter */
		    matched = TRUE;
		    cs = MNEXT(cs);
		} else {
		    DEBUG_RETURN(12,NIL);
		    /* no match, error return */
		    return (NIL);
		}
		break;

	    /* check for start of line */
	    case '^':
		if (s == Str_Start) {

		    /* match, be happy */
		    matched = TRUE;
		    cs = MNEXT(cs);
		} else if (*cs & ALT) {

		    /* try the next part */
		    matched = FALSE;
		    cs = MNEXT(cs);
		} else if (*cs & OPT) {

		    /* doesn't matter */
		    matched = TRUE;
		    cs = MNEXT(cs);
		} else {
		    DEBUG_RETURN(13,NIL);
		    /* no match, error return */
		    return (NIL);
		}
		break;

	    /* end of a subexpression, return success */
	    case ')':
		DEBUG_RETURN(105,s);
		return (s);
	    }
	    break;
	}
    }
    DEBUG_RETURN(106,s);
    return (s);
}

#if STDC
void
set_isidchar(const char *extra)
#else
void
set_isidchar(extra)
const char *extra;
#endif
{
    int k;

    /* The newest vgrind implementations (e.g. Sun Solaris 2.x) */
    /* allow :id=....: to extend the set of identifier characters */
    /* that match \p.  Unfortunately, the basic set A-Za-z0-9_ is */
    /* hard coded, and can only be extended, not replaced. */

    for (k = 0; k < elementsof(isidchar); ++k)
	isidchar[k] = (isalnum(k) || (k == (int)'_')) ? TRUE : FALSE;

    if (extra)
    {
	for (k = 0; extra[k]; ++k)
	    isidchar[CHRVAL(extra[k])] = TRUE;
    }
}

#if STDC
void
set_isinichar(const char *extra)
#else
void
set_isinichar(extra)
const char *extra;
#endif
{
    int k;

    if (extra)
    {
	for (k = 0; k < elementsof(isinichar); ++k)
	    isinichar[k] = (isalnum(k) || (k == (int)'_')) ? TRUE : FALSE;

	for (k = 0; extra[k]; ++k)
	    isinichar[CHRVAL(extra[k])] = TRUE;
    }
    else			/* default is idchar set */
    {
	for (k = 0; k < elementsof(isinichar); ++k)
	    isinichar[k] = isidchar[k];
    }
}

#if STDC
void
unset_isidchar(const char *remove_chars)
#else
void
unset_isidchar(remove_chars)
const char *remove_chars;
#endif
{
    int k;

    /* Remove characters from the set of identifier characters, so */
    /* that a language specification can work around the hard coding */
    /* of the basic identifier set A-Za-z0-9_. */

    if (remove_chars)
    {
	for (k = 0; remove_chars[k]; ++k)
	    isidchar[CHRVAL(remove_chars[k])] = FALSE;
    }
}

#if STDC
void
unset_isinichar(const char *remove_chars)
#else
void
unset_isinichar(remove_chars)
const char *remove_chars;
#endif
{
    int k;

    /* Remove characters from the set of initial identifier characters, so */
    /* that a language specification can work around the hard coding */
    /* of the basic identifier set A-Za-z0-9_. */

    if (remove_chars)
    {
	for (k = 0; remove_chars[k]; ++k)
	    isinichar[CHRVAL(remove_chars[k])] = FALSE;
    }

}
